@@ -5,10 +5,10 @@ from functools import partial
 import torch.nn.functional as F
 
 from math import sqrt
-from einops import rearrange, repeat
+from einops import rearrange, repeat, reduce
 
 import alphafold2_pytorch.constants as constants
-from alphafold2_pytorch.reversible import ReversibleSequence, SequentialSequence
+from alphafold2_pytorch.reversible import ReversibleSequence
 
 from se3_transformer_pytorch import SE3Transformer
 
@@ -22,6 +22,9 @@ def default(val, d):
         return val
     return d() if isfunction(d) else d
 
+def cast_tuple(val, depth):
+    return val if isinstance(val, tuple) else (val,) * depth
+
 # helper classes
 
 class PreNorm(nn.Module):
@@ -75,12 +78,15 @@ class Attention(nn.Module):
     def __init__(
         self,
         dim,
+        seq_len = None,
         heads = 8,
         dim_head = 64,
-        dropout = 0.
+        dropout = 0.,
+        compress_ratio = 1,
     ):
         super().__init__()
         inner_dim = dim_head * heads
+        self.seq_len = seq_len
         self.heads= heads
         self.scale = dim_head ** -0.5
         self.to_q = nn.Linear(dim, inner_dim, bias = False)
@@ -89,18 +95,46 @@ class Attention(nn.Module):
 
         self.dropout = nn.Dropout(dropout)
 
+        # memory compressed attention
+        self.compress_ratio = compress_ratio
+        self.compress_fn = nn.Conv1d(inner_dim, inner_dim, compress_ratio, stride = compress_ratio, groups = heads) if compress_ratio > 1 else None
+
     def forward(self, x, context = None, mask = None, context_mask = None):
         device, h, has_context = x.device, self.heads, exists(context)
         context = default(context, x)
 
         q, k, v = (self.to_q(x), *self.to_kv(context).chunk(2, dim = -1))
+
+        i, j = q.shape[-2], k.shape[-2]
+
+        if exists(self.compress_fn):
+            assert has_context, 'memory compressed attention only works in the context of cross attention for now'
+
+            ratio = self.compress_ratio
+            padding = ratio - (j % ratio)
+
+            if padding < ratio:
+                k, v = map(lambda t: F.pad(t, (0, 0, 0, padding), value = 0), (k ,v))
+
+                if exists(context_mask):
+                    context_mask = F.pad(context_mask, (0, padding), value = False)
+
+                k, v = map(lambda t: rearrange(t, 'b n c -> b c n'), (k, v))
+                k, v = map(self.compress_fn, (k, v))
+                k, v = map(lambda t: rearrange(t, 'b c n -> b n c'), (k, v))
+
+                context_mask = reduce(context_mask.float(), 'b (n r) -> b n', 'sum', r = ratio)
+                context_mask = context_mask > 0
+
+                j = (j + padding) // ratio
+
         q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))
 
         dots = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale
 
         if exists(mask) or exists(context_mask):
-            mask = default(mask, lambda: torch.ones(*x.shape[:2], device = device))
-            context_mask = default(context_mask, mask) if not has_context else default(context_mask, lambda: torch.ones(*context.shape[:2], device = device))
+            mask = default(mask, lambda: torch.ones(i, device = device))
+            context_mask = default(context_mask, mask) if not has_context else default(context_mask, lambda: torch.ones(j, device = device))
             mask_value = -torch.finfo(dots.dtype).max
             mask = mask[:, None, :, None] * context_mask[:, None, None, :]
             dots.masked_fill_(~mask, mask_value)
@@ -112,14 +146,66 @@ class Attention(nn.Module):
         out = rearrange(out, 'b h n d -> b n (h d)')
         return self.to_out(out)
 
+class SparseAttention(Attention):
+    def __init__(
+        self,
+        *args,
+        block_size = 16,
+        num_random_blocks = None,
+        **kwargs
+    ):
+        super().__init__(*args, **kwargs)
+        assert exists(self.seq_len), '`seq_len` must be defined if using sparse attention class'
+        from deepspeed.ops.sparse_attention import SparseSelfAttention, VariableSparsityConfig
+        self.block_size = block_size
+        num_random_blocks = default(num_random_blocks, self.seq_len // block_size // 4)
+
+        self.attn_fn = SparseSelfAttention(
+            sparsity_config = VariableSparsityConfig(
+                num_heads = self.heads,
+                block = self.block_size,
+                num_random_blocks = num_random_blocks,
+                attention = 'bidirectional'
+            ),
+            max_seq_length = self.seq_len,
+            attn_mask_mode = 'add'
+        )
+
+    def forward(self, x, mask = None):
+        device, h = x.device, self.heads
+        n = x.shape[1]
+
+        remainder = x.shape[1] % self.block_size
+
+        if remainder > 0:
+            padding = self.block_size - remainder
+            x = F.pad(x, (0, 0, 0, padding), value = 0)
+            mask = F.pad(mask, (0, padding), value = False)
+
+        q, k, v = (self.to_q(x), *self.to_kv(x).chunk(2, dim = -1))
+        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))
+
+        dots = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale
+
+        key_pad_mask = None
+        if exists(mask):
+            key_pad_mask = ~mask
+
+        out = self.attn_fn(q, k, v, key_padding_mask = key_pad_mask)
+        out = rearrange(out, 'b h n d -> b n (h d)')
+        out = self.to_out(out)
+        return out[:, :n]
+
 class AxialAttention(nn.Module):
     def __init__(
         self,
+        sparse_attn = False,
         **kwargs
     ):
         super().__init__()
-        self.attn_width = Attention(**kwargs)
-        self.attn_height = Attention(**kwargs)
+        attn_class = SparseAttention if sparse_attn else Attention
+        self.attn_width = attn_class(**kwargs)
+        self.attn_height = attn_class(**kwargs)
 
     def forward(self, x, context = None, mask = None, context_mask = None):
         b, n, d = x.shape
@@ -140,12 +226,13 @@ class AxialAttention(nn.Module):
             w_context_mask = repeat(context_mask, 'b n -> (b w) n', w = w)
             h_context_mask = repeat(context_mask, 'b n -> (b h) n', h = h)
 
+        attn_kwargs = {} if not exists(context) else {'context': w_context, 'context_mask': w_context_mask}
         w_x = rearrange(x, 'b h w d -> (b w) h d')
-        w_out = self.attn_width(w_x, mask = w_mask, context = w_context, context_mask = w_context_mask)
+        w_out = self.attn_width(w_x, mask = w_mask, **attn_kwargs)
         w_out = rearrange(w_out, '(b w) h d -> b h w d', h = h, w = w)
 
         h_x = rearrange(x, 'b h w d -> (b h) w d')
-        h_out = self.attn_height(h_x, mask = h_mask, context = h_context, context_mask = h_context_mask)
+        h_out = self.attn_height(h_x, mask = h_mask, **attn_kwargs)
         h_out = rearrange(h_out, '(b h) w d -> b h w d', h = h, w = w)
 
         out = w_out + h_out
@@ -153,6 +240,35 @@ class AxialAttention(nn.Module):
 
 # main class
 
+class SequentialSequence(nn.Module):
+    def __init__(self, blocks):
+        super().__init__()
+        self.blocks = blocks
+
+    def forward(self, x, m, mask = None,  msa_mask = None, **kwargs):
+        for ((attn, ff, msa_attn), (cross_attn, msa_ff, msa_cross_attn)) in zip(*[iter(self.blocks)] * 2):
+
+            # self attention
+
+            x = attn(x, mask = mask) + x
+
+            if exists(m):
+                m = msa_attn(m, mask = msa_mask) + m
+
+                # cross attention
+
+                x = cross_attn(x, m, mask = mask, context_mask = msa_mask) + x
+                m = msa_cross_attn(m, x, mask = msa_mask, context_mask = mask) + m
+
+            # feedforwards
+
+            x = ff(x) + x
+
+            if exists(m):
+                m = msa_ff(m) + m
+
+        return x, m
+
 class Alphafold2(nn.Module):
     def __init__(
         self,
@@ -167,11 +283,16 @@ class Alphafold2(nn.Module):
         num_embedds = constants.NUM_EMBEDDS_TR,
         attn_dropout = 0.,
         ff_dropout = 0.,
-        reversible = False
+        reversible = False,
+        sparse_self_attn = False,
+        cross_attn_compress_ratio = 1,
     ):
         super().__init__()
+        layers_sparse_attn = cast_tuple(sparse_self_attn, depth)
+
         self.token_emb = nn.Embedding(num_tokens, dim)
         self.pos_emb = nn.Embedding(max_seq_len, dim)
+        self.pos_emb_ax = nn.Embedding(max_seq_len, dim)
 
         # multiple sequence alignment position embedding
 
@@ -188,23 +309,25 @@ class Alphafold2(nn.Module):
         prenorm_cross = partial(PreNormCross, dim)
 
         layers = nn.ModuleList([])
-        for _ in range(depth):
+        for _, layer_sparse_attn in zip(range(depth), layers_sparse_attn):
 
             # self attention, for both main sequence and msa
 
+            attn_class = SparseAttention if layer_sparse_attn else Attention
+
             layers.append(nn.ModuleList([
-                prenorm(AxialAttention(dim = dim, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
+                prenorm(AxialAttention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, sparse_attn = sparse_self_attn)),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
-                prenorm(Attention(dim = dim, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
+                prenorm(attn_class(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
             ]))
 
             # cross attention, for main sequence -> msa and then msa -> sequence
 
             layers.append(nn.ModuleList([
-                prenorm_cross(Attention(dim = dim, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
+                prenorm_cross(Attention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, compress_ratio = cross_attn_compress_ratio)),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
-                prenorm_cross(Attention(dim = dim, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
+                prenorm_cross(Attention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, compress_ratio = cross_attn_compress_ratio)),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
             ]))
 
@@ -232,9 +355,17 @@ class Alphafold2(nn.Module):
         # embed main sequence
 
         x = self.token_emb(seq)
-        x += self.pos_emb(torch.arange(n, device = device))[None, ...]
-        x = x[:, :, None, :] + x[:, None, :, :] # create pair-wise residue embeds
-        x_mask = mask[:, :, None] * mask[:, None, :]
+
+        # use axial positional embedding
+
+        seq_range = torch.arange(n, device = device)
+        ax1 = x + self.pos_emb(seq_range)[None, ...]
+        ax2 = x + self.pos_emb_ax(seq_range)[None, ...]
+
+        # outer sum
+
+        x = rearrange(ax1, 'b i d -> b i () d') + rearrange(ax2, 'b j d-> b () j d') # create pair-wise residue embeds
+        x_mask = rearrange(mask, 'b i -> b i ()') + rearrange(mask, 'b j -> b () j')
 
         x = rearrange(x, 'b i j d -> b (i j) d')
         x_mask = rearrange(x_mask, 'b i j -> b (i j)')
@@ -253,7 +384,7 @@ class Alphafold2(nn.Module):
 
         elif exists(embedds):
             m = self.embedd_project(embedds)
-            m = m[:, :, None, :] + m[:, None, :, :]
+            m = rearrange(m, 'b i d -> b i () d') + rearrange(m, 'b j d -> b () j d')
             m = rearrange(m, 'b m n d -> b (m n) d')
 
         if exists(msa_mask):
